#include <iostream>
#include "fcos3d_pipeline.h"
#include "fcos_parameters.h"
#include "utils.h"
using namespace std;
bool g_isDevice = false;

#include <vector>
#include <cmath>
#include <numeric>
#include <algorithm>
// #include <preprocess.h>
// #include "fcos_base_pipeline.h"
// #include "fcos_base_parameters.h"
#include "utils.h"
#include <sys/time.h>
// #include "runModel.h"
// #include "postprocess.h"
#include "convert_3dbbox_to_2bbox.h"
#include "extractCanData.h"

#include "lidar_tracker.h"
#include "MOT_Wrapper.h"
// #define Wrapper_Operation
#ifdef Wrapper_Operation
#include "ara/log/logging.h"
#include "LidarPerception_GeneralWrapper.h"
#endif // Wrapper_Operation

#include <thread>
// #include "data_reader.h"

#include "compute_ENU.h"
#include "transfer_to_ECEF.h"
#include "print_Ins.h"

#include <cmath>

#include <opencv2/core.hpp>
#include <opencv2/highgui.hpp>
#include <opencv2/imgproc.hpp>
#include <opencv2/videoio.hpp>

#include <Eigen/Dense>

using namespace lidar_tracker;
ObservationsProto BuildObservation(const std::vector<Box2d> &boxes,
                                   const std::vector<int> &ids,
                                   const std::vector<ObservationType> types,
                                   const double timestamp)
{
    ObservationsProto observation;
    observation.set_min_timestamp(timestamp);
    for (int i = 0; i < boxes.size(); ++i)
    {
        auto detection = observation.add_observations();
        detection->set_timestamp(timestamp);
        detection->set_id(ids[i]);
        detection->set_type(types[i]);

        auto lidar_observation = detection->mutable_lidar_observation();
        lidar_observation->set_timestamp(timestamp);
        lidar_observation->set_center_x(boxes[i].center_x());
        lidar_observation->set_center_y(boxes[i].center_y());
        lidar_observation->set_length(boxes[i].length());
        lidar_observation->set_width(boxes[i].width());
        lidar_observation->set_yaw(boxes[i].heading());
        lidar_observation->set_type(types[i]);
    }
    return observation;
}
ObservationsProto frame;
std::vector<int> ids;
std::vector<ObservationType> types;
// std::vector<double> timestamp;
std::vector<Box2d> LidarObservation;
double timestamp;
std::vector<perception::LiDARPerception2Fusion> MOT_output_objs;
// ara::log::Logger& mainLog{ara::log::CreateLogger("main", "Lidar MOT data logging", ara::log::LogLevel::kVerbose)};
// mainLog.LogInfo() << "Lidar_MOT module data...";
// GeneralWrapper gw(mainLog);
dpc::SLAMLocalEgoOdom vehicle_odom_vec_Temp;
std::vector<LidarPerceptionOutput> RuleBasedResult;
mot::InputWrapper wrapper(50); // 50: buffer size for ego motion interpollation
LidarTracker tracker;
bool Loop_finished = false;
int package_counter = 0;

int start_frame = 0; //0; // 270; // 784; // 987;  //1131;
int end_frame = 15; //19;  // 275;   // 288; // 805;   //1032;  //1131 + 50;
const double PI = 3.14159265358979323846;

#ifdef Wrapper_Operation
ara::log::Logger &mainLog{ara::log::CreateLogger("main", "Lidar MOT data logging", ara::log::LogLevel::kVerbose)};
mainLog.LogInfo() << "Lidar_MOT module data...";
GeneralWrapper gw(mainLog);
#endif // Wrapper_Operation

int main()
{
    Eigen::Matrix4f cam2img;
    const std::string txtfilename = "./camera_8_messages.txt";
    std::vector<double> timestampsHigh;
    std::vector<double> timestampsLow;
    // Call the function to read timestamps
    readTimestamps(txtfilename, timestampsHigh, timestampsLow);

    // extract the localization outputs from the CAN data file(can_data_messages.txt file)
    std::string filename = "./can_data_messages.txt";
    Eigen::MatrixXf speed_matrix;
    Eigen::MatrixXf ang_speed_matrix;
    Eigen::MatrixXf yaw_rate_matrix;
    Eigen::MatrixXf position_z_matrix, position_y_matrix, position_x_matrix;
    Eigen::MatrixXf acc_matrix = extractAccelerationData(filename);
    extractPositionData(filename, position_z_matrix, position_y_matrix, position_x_matrix);
    Eigen::MatrixXf orientation_matrix = extractOrientationData(filename);
    extractSpeedAndYawRateData(filename, speed_matrix, yaw_rate_matrix, ang_speed_matrix);
    std::vector<std::tuple<float, float, float, long long, long long>> time_stamp_data = extractAccData(filename);

    std::vector<Eigen::Vector3d> enu_points;
    std::vector<Eigen::Vector3d> points;
    std::vector<double> e_coordinates, n_coordinates;
    fillPoints(points, position_x_matrix, position_y_matrix, start_frame, end_frame);
    print_INS(points, enu_points);

    // Extract E and N coordinates from enu_points
    extractMetricCoordinates(enu_points, e_coordinates, n_coordinates);

    cv::Size imageSize(1600, 900);
    int frameCount = 30; // Number of frames in the video
    int fps = 30;        // Frames per second

    // Create a VideoWriter object
    cv::VideoWriter videoWriter("/opt/usr/naraghi/output.avi", cv::VideoWriter::fourcc('M', 'J', 'P', 'G'), fps, imageSize);

    if (!videoWriter.isOpened())
    {
        std::cerr << "Error: Could not open the video file for writing" << std::endl;
        return -1;
    }

    Fcos3dPipeline fcos3DPipeline;
    Fcos3dInitParams fcos3DInitParams("./opt/config/config.yaml");
    WARN_LOG("Resource initialization begin");
    Result ret = fcos3DPipeline.Init(fcos3DInitParams);
    if (ret != SUCCESS)
    {
        ERROR_LOG("Resource initialization failed");
        return FAILED;
    }
    INFO_LOG("Resource initialization finished");

    WARN_LOG("Model processing begin");

    ret = fcos3DPipeline.Process();
    if (ret != SUCCESS)
    {
        ERROR_LOG("Model processing failed");
        return FAILED;
    }
    // }
    INFO_LOG("Model processing finished");
    std::vector<float> f;

    for (int i = start_frame; i < end_frame; i++) // 458   789 < 800
    {
        // set odometry_data
        vehicle_odom_vec_Temp.ego_odom_time_ns = std::get<3>(time_stamp_data[i - 549]) + (std::get<4>(time_stamp_data[i - 549])) / 1e9;
        // vehicle_odom_vec_Temp.ego_odom_time_ns = timestampsHigh[i] + (timestampsLow[i]/1e9);

        vehicle_odom_vec_Temp.position.x = e_coordinates[int(i)]; // position_x_matrix(i, 0); *19.54/20.35
        vehicle_odom_vec_Temp.position.y = n_coordinates[int(i)]; // position_y_matrix(i, 0);
        vehicle_odom_vec_Temp.position.z = position_z_matrix(int(i), 0);

        vehicle_odom_vec_Temp.roll = orientation_matrix(int(i), 2);  // x
        vehicle_odom_vec_Temp.pitch = orientation_matrix(int(i), 1); // y
        vehicle_odom_vec_Temp.yaw = orientation_matrix(int(i), 0);   // z

        vehicle_odom_vec_Temp.linear_velocity.x = speed_matrix(int(i), 1);
        vehicle_odom_vec_Temp.linear_velocity.y = speed_matrix(int(i), 2);
        vehicle_odom_vec_Temp.linear_velocity.z = speed_matrix(int(i), 3);

        vehicle_odom_vec_Temp.angular_velocity.x = ang_speed_matrix(int(i), 0);
        vehicle_odom_vec_Temp.angular_velocity.y = ang_speed_matrix(int(i), 1);
        vehicle_odom_vec_Temp.angular_velocity.z = ang_speed_matrix(int(i), 2);

        vehicle_odom_vec_Temp.linear_acceleration.x = acc_matrix(int(i), 0);
        vehicle_odom_vec_Temp.linear_acceleration.y = acc_matrix(int(i), 1);

        std::cout << "frameInd : " << i << std::endl;
        struct timeval start_time;
        struct timeval end_time;
        gettimeofday(&start_time, NULL);

        std::string nameImg = "CAM_BACK" + std::to_string(i + 1) + ".jpg";
        // std::string nameImg = "undistorted_frame_" + std::to_string(i+270) + ".png";

        cv::Mat image;
        image = cv::imread("./Nuscenes_CAM_BACK/" + nameImg);
        // image = cv::imread("./undistorted_frame_27/" + nameImg);

        std::vector<PredOut<float>> allResults = fcos3DPipeline.GetAllResults();
        std::cout << " main : size allResults : " << allResults.size() << std::endl;
        Eigen::MatrixXf bboxes_3d = allResults[i].bboxes;
        Eigen::VectorX<float> scores = allResults[i].scores;
        Eigen::VectorXi labels = allResults[i].labels;

        // filter the objects base on the scores > 0.1 and save their index in the best_object
        std::vector<int> best_object;
        for (int num = 0; num < scores.size(); ++num)
        {
            if (scores[num] > 0.2)
            {
                best_object.push_back(num);
            }
        }

        // *****     convert 3d to 2d  *********   //
        std::vector<Eigen::MatrixXf> corners_3d = BBox3DTo2DConverter2::corners(bboxes_3d);
        // cam2img << 1168.5819220882242 * 0.8333, 0.0, 962.77736854943203 * 0.8333, 0,
        //     0.0, 1168.413278634928 * 0.8333, 544.3403960076156 * 0.8333, 0,
        //     0.0, 0.0, 1.0, 0,
        //     0, 0, 0, 1;

        cam2img << 809.221, 0.0, 829.2196, 0,
            0.0, 809.221, 481.7784, 0,
            0.0, 0.0, 1.0, 0,
            0, 0, 0, 1;

        std::vector<Eigen::MatrixXf> corners_2d = BBox3DTo2DConverter2::proj_camera_bbox3d_to_img(corners_3d, cam2img);

        // Call the function to calculate [min-x max-x min-y max-y] of bboxes and plot
        Eigen::MatrixXf bbox_2d = BBox3DTo2DConverter2::calculateMinMaxPlotRectangles(image, corners_2d);

#ifdef Wrapper_Operation
        gw._SLAMLocalEgoOdom2SWC_Required.s.GetData(vehicle_odom_vec_Temp);
#endif // Wrapper_Operation

        // Doing the coordinate conversion
        timestamp = timestampsHigh[i] + (timestampsLow[i] / 1e9);
        wrapper.MotWrapper(bboxes_3d, best_object, vehicle_odom_vec_Temp, timestamp);

        // Iterate over all objects (read-only)
        uint32_t numOfObj = 0;
        for (int num = 0; num < best_object.size(); ++num)
        {
            ids.push_back(numOfObj++);
            // types.push_back(obj.type);
            types.push_back(static_cast<ObservationType>(labels[best_object[num]]));

            double center_x = double(bbox_2d(best_object[num], 0) + bbox_2d(best_object[num], 1)) / 2;
            double center_y = double(bbox_2d(best_object[num], 2) + bbox_2d(best_object[num], 3)) / 2;
            double height = bbox_2d(best_object[num], 3) - bbox_2d(best_object[num], 2);
            double width = bbox_2d(best_object[num], 1) - bbox_2d(best_object[num], 0);

            Vec2d v(center_x, center_y);
            Box2d b(v, double(bboxes_3d(best_object[num], 6) * (PI / 180)), height, width);
            // LidarObservation.push_back = { {double(bboxes_3d(best_object[num], 0)), double(bboxes_3d(best_object[num], 1))}, double(bboxes_3d(best_object[num], 6)), double(bboxes_3d(best_object[num], 6)), double(bboxes_3d(best_object[num], 6)) };
            LidarObservation.push_back(b);
        }

        frame = BuildObservation(LidarObservation, ids, types, timestamp);

        tracker.Tracking(frame);
        // Check result
        auto tracks = tracker.GetTracks();
        tracker.GetTracks().size();
        std::cout << "size of obstacles : " << tracker.GetTracks().size() << std::endl;
        output_wrapper(MOT_output_objs, tracks);
        // Doing the coordinate conversion
        Kinematics_local2overground(MOT_output_objs, &vehicle_odom_vec_Temp);

#ifdef Wrapper_Operation
        gw._LiDARPerception2Fusion_Provide.s.SetData(MOT_output_objs);
        gw._LiDARPerception2Fusion_Provide.Publish();
#endif // Wrapper_Operation

        Eigen::MatrixXf min_max(tracker.GetTracks().size(), 4);

        for (int j = 0; j < tracker.GetTracks().size(); ++j)
        {
            double min_x = tracker.GetTracks().at(j)->track_state.bounding_box.center_x() -
                           tracker.GetTracks().at(j)->track_state.bounding_box.half_width();
            double max_x = tracker.GetTracks().at(j)->track_state.bounding_box.center_x() +
                           tracker.GetTracks().at(j)->track_state.bounding_box.half_width();
            double min_y = tracker.GetTracks().at(j)->track_state.bounding_box.center_y() -
                           tracker.GetTracks().at(j)->track_state.bounding_box.half_length();
            double max_y = tracker.GetTracks().at(j)->track_state.bounding_box.center_y() +
                           tracker.GetTracks().at(j)->track_state.bounding_box.half_length();

            min_max(j, 0) = min_x;
            min_max(j, 1) = max_x;
            min_max(j, 2) = min_y;
            min_max(j, 3) = max_y;

            // Create rectangle
            cv::Rect rect(min_x, min_y, max_x - min_x, max_y - min_y);

            // // Create rectangle
            // cv::Rect rect(min_x, min_y, max_y - min_y, max_x - min_x);

            // Draw rectangle on image
            cv::rectangle(image, rect, cv::Scalar(0, 255, 0), 2);

            // Write the frame to the video
            videoWriter.write(image);

            // save the frame as a JPEG file
            std::string picname = "frame_" + std::to_string(i) + ".jpg";
            cv::imwrite("/opt/usr/naraghi/" + picname, image);
        }
        // // Display the image with rectangles
        // cv::imshow("Image with Rectangles", image);
        // cv::waitKey(0);

        LidarObservation.clear();
        MOT_output_objs.clear();
        ids.clear();
        types.clear();
    }
    // Release the video writer
    videoWriter.release();

    std::cout << "Video has been written successfully :))" << std::endl;

    return SUCCESS;
}